package net.astah.stpa.stamp.notation.edit.command;

import static java.util.Collections.singleton;
import static net.astah.dd.di.util.DiagramElements.getModelElement;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.IdentityCommand;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.ecore.EObjects;
import net.astah.emf.edit.command.DeferringCommandWrapper;
import net.astah.golf.util.RuntimeGlobal;
import net.astah.notation.NotationFactory;
import net.astah.notation.NotationPackage;
import net.astah.notation.Shape;
import net.astah.notation.Size;
import net.astah.notation.edit.command.ApplyPreferenceStyleCommand;
import net.astah.project.Facet;
import net.astah.project.Project;
import net.astah.project.ProjectManagerAdapter;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ControlLink;
import net.astah.stpa.stamp.ControlStructure;
import net.astah.stpa.stamp.Element;
import net.astah.stpa.stamp.FeedbackLink;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.LinkEnd;
import net.astah.stpa.stamp.Note;
import net.astah.stpa.stamp.NoteAnchor;
import net.astah.stpa.stamp.STAMPPackage;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.Signal;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.notation.STAMPControlStructureDiagram;
import net.astah.stpa.stamp.notation.STAMPDiagram;
import net.astah.stpa.stamp.notation.STAMPDiagramElement;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.STAMPNotationPackage;
import net.astah.stpa.stamp.notation.STAMPShape;
import net.astah.stpa.stamp.notation.util.STAMPNotationSwitch;
import net.astah.stpa.stamp.util.STAMPSwitch;

/**
 * 図要素モデルが図に登録される際の初期化処理を実装します。
 */
public class InitializeSTAMPDiagramElementCommand extends CompoundCommand {

    static final Logger LOGGER = LoggerFactory
            .getLogger(InitializeSTAMPDiagramElementCommand.class);

    /**
     * コマンドを実行するドメイン
     */
    protected EditingDomain domain;

    /**
     * 登録対象の図
     */
    protected EObject owner;

    /**
     * 登録対象の図要素モデル
     */
    protected Object value;

    /**
     * 登録する位置
     */
    protected int index;

    /**
     * コンストラクタ
     * 
     * @param domain
     *            コマンドを実行するドメイン
     * @param owner
     *            登録対象の図
     * @param value
     *            登録対象の図要素モデル
     * @param index
     *            登録する位置
     */
    public InitializeSTAMPDiagramElementCommand(final EditingDomain domain, final EObject owner,
            final Object value, int index) {
        super("InitializeSTAMPDiagramElementCommand", "InitializeSTAMPDiagramElementCommand");
        this.domain = domain;
        this.owner = owner;
        this.value = value;
        this.index = index;
    }

    @Override
    protected boolean prepare() {

        if (owner == null || value == null) {
            return false;
        }

        final STAMPDiagramElement diagramElement = (STAMPDiagramElement) value;
        final EList<Element> modelElements = diagramElement.getSTAMPModelElements();
        if (!modelElements.isEmpty()) {
            for (final Element element : modelElements) {
                // 対象のモデル要素がドメインの制御下にない場合に新規登録と見なす
                if (!EObjects.isAttachedTo(domain, element)) {
                    append(createCreateChildCommand(element));
                }
            }
        }

        // 初期化処理(図要素モデルから配下のラベルを作成する処理を含む)
        AddInitializeCommand(diagramElement);

        return isEmpty() || super.prepare();
    }

    /**
     * モデル要素を新規登録するコマンドを作成します。
     * 
     * @param element
     *            新規登録されるモデル要素
     * @return 作成されたコマンド
     */
    protected Command createCreateChildCommand(final Element element) {
        final NewChildDescriptorFactory factory = new NewChildDescriptorFactory();
        final CommandParameter newChildDescriptor = factory.doSwitch(element);
        if (newChildDescriptor == null) {
            return UnexecutableCommand.INSTANCE;
        } else if (newChildDescriptor.getOwner() == null && newChildDescriptor.getFeature() == null
                && newChildDescriptor.getValue() == null) {
            return IdentityCommand.INSTANCE;
        }
        final Object owner = newChildDescriptor.getOwner();
        return CreateChildCommand.create(domain, owner, newChildDescriptor, singleton(owner));
    }

    /**
     * 初期化処理のコマンドを作成します。
     * 
     * @param diagramElement
     *            対象の図要素モデル
     */
    protected void AddInitializeCommand(final STAMPDiagramElement diagramElement) {
        new AddInitializeCommandSwitch().doSwitch(diagramElement);
    }

    /**
     * STPA プロジェクトの {@link Facet} を返します。
     * 
     * @return STPA プロジェクトの {@link Facet}
     * @see STAMPPlugin#FACET_SYMBOLIC_NAME
     */
    protected Facet getFacet() {
        final ResourceSet resourceSet = domain.getResourceSet();
        final Project project = ProjectManagerAdapter.getContainingProject(resourceSet, owner);
        return project != null ? project.getFacet(STAMPPlugin.FACET_SYMBOLIC_NAME) : null;
    }

    /**
     * モデル要素を新規登録するための {@link CommandParameter} を作成するクラスです。
     */
    protected class NewChildDescriptorFactory extends STAMPSwitch<CommandParameter> {

        @Override
        public CommandParameter caseComponent(final Component object) {
            if (owner instanceof STAMPComponentShape) {
                return new CommandParameter(getModelElement((STAMPComponentShape) owner),
                        STAMPPackage.Literals.COMPONENT__NESTED_COMPONENT, object, index);
            }
            return getCommandParameter(this, STAMPPackage.Literals.CONTROL_STRUCTURE__COMPONENT,
                    object);
        }

        @Override
        public CommandParameter caseLink(final Link object) {
            return getCommandParameter(this, STAMPPackage.Literals.CONTROL_STRUCTURE__LINK,
                    object);
        }

        @Override
        public CommandParameter caseSignal(final Signal object) {
            return new CommandParameter(getModelElement((STAMPEdge) owner),
                    STAMPPackage.Literals.LINK__SIGNAL, object, index);
        }

        @Override
        public CommandParameter caseNote(final Note object) {
            return getCommandParameter(this, STAMPPackage.Literals.ELEMENT__OWNED_NOTE,
                    object);
        }

        @Override
        public CommandParameter caseLinkEnd(final LinkEnd object) {
            return getCommandParameter(this, STAMPPackage.Literals.CONTROL_STRUCTURE__LINK_END,
                    object);
        }

        @Override
        public CommandParameter caseNoteAnchor(final NoteAnchor object) {
            // ノートアンカーのownerはSTAMPEdgeItemProvider.createSetEdgeEndCommand()で設定するのでここでは何もしない
            return new CommandParameter(null, null, null);
        }
    }

    private CommandParameter getCommandParameter(
            NewChildDescriptorFactory newChildDescriptorFactory,
            EReference eReference, EObject eObject) {
        final STAMPDiagram diagram = EObjects.getAncestor(STAMPDiagram.class, owner);
        if (diagram instanceof STAMPControlStructureDiagram) {
            for (final Element element : diagram.getSTAMPModelElements()) {
                if (element instanceof ControlStructure) {
                    return new CommandParameter(element, eReference, eObject);
                }
            }
        } else if (diagram instanceof STAMPControlLoopDiagram) {
            final STPAAnalysis analysis = ((STAMPControlLoopDiagram) diagram).getControlLoop()
                    .getHintWordSet().getOwner();
            final ControlStructure controlStructure = analysis.getControlStructures().stream()
                    .findFirst().orElseGet(analysis::createControlStructure);
            return new CommandParameter(controlStructure, eReference, eObject);
        }
        return newChildDescriptorFactory.doSwitch(eObject);
    }

    /**
     * 初期化処理のコマンドを作成するクラスです。
     */
    protected class AddInitializeCommandSwitch extends STAMPNotationSwitch<Void> {

        @Override
        public Void caseDiagramElement(DiagramElement object) {
            final String sharedFontKey = "font";
            final Project project = ProjectManagerAdapter.getRootProject(owner);
            append(new DeferringCommandWrapper(ApplyPreferenceStyleCommand.create(domain, object,
                    project, sharedFontKey)));
            return super.caseDiagramElement(object);
        }

        @Override
        public Void caseSTAMPComponentShape(final STAMPComponentShape object) {
            if (!object.isOmitted()) {
                final STAMPLabel label = STAMPNotationFactory.eINSTANCE.createSTAMPLabel();
                EObject model = getModelElement(object);
                DiagramElements.setModelElement(label, model);

                append(AddCommand.create(domain, object,
                        STAMPNotationPackage.Literals.STAMP_DIAGRAM_ELEMENT__STAMP_OWNED_ELEMENT,
                        label));
                // コンポーネントの図要素が作成された場合、モデルの分析対象のフラグをtrueに設定する
                append(SetCommand.create(domain, model,
                        STAMPPackage.Literals.COMPONENT__IS_ANALYSIS_TARGET, true));
                append(new SynchronizeLabelCommand(domain, label));

                // システムプロパティのデフォルトサイズが設定されている場合、サイズを指定する
                appendSetDefaultSizeIfAbsent(object, "component");

                // 共通スタイルの適用
                final String styleKey = "component";
                append(ApplyPreferenceStyleCommand.create(domain, object, getFacet(), styleKey));
            }
            return super.caseSTAMPComponentShape(object);
        }

        @Override
        public Void caseSTAMPShape(final STAMPShape object) {
            final EObject model = getModelElement(object);
            if (model.eClass() == STAMPPackage.Literals.NOTE) {
                final STAMPLabel label = STAMPNotationFactory.eINSTANCE.createSTAMPLabel();
                append(AddCommand.create(domain, object,
                        STAMPNotationPackage.Literals.STAMP_DIAGRAM_ELEMENT__STAMP_OWNED_ELEMENT,
                        label));
                append(AddCommand.create(domain, label,
                        STAMPNotationPackage.Literals.STAMP_DIAGRAM_ELEMENT__STAMP_MODEL_ELEMENT,
                        object.getSTAMPModelElements()));
                // システムプロパティのデフォルトサイズが設定されている場合、サイズを指定する
                appendSetDefaultSizeIfAbsent(object, "note");

                // 共通スタイルの適用
                final String styleKey = "note";
                append(ApplyPreferenceStyleCommand.create(domain, object, getFacet(), styleKey));
            }
            return super.caseSTAMPShape(object);
        }

        @Override
        public Void caseSTAMPEdge(final STAMPEdge object) {
            final EObject modelElement = getModelElement(object);
            if (modelElement instanceof ControlLink) {
                // 共通スタイルの適用
                final String styleKey = "control_action";
                append(ApplyPreferenceStyleCommand.create(domain, object, getFacet(), styleKey));
            } else if (modelElement instanceof FeedbackLink) {
                // 共通スタイルの適用
                final String styleKey = "feedback";
                append(ApplyPreferenceStyleCommand.create(domain, object, getFacet(), styleKey));
            }
            return super.caseSTAMPEdge(object);
        }

        private void appendSetDefaultSizeIfAbsent(final Object object, final String kind) {
            if (!(object instanceof Shape)) {
                return;
            }
            final Shape shape = (Shape) object;
            if (shape.getSize() != null) {
                return;
            }
            final String defaultKey = "default_size." + kind;
            final String widthKey = defaultKey + ".default_width";
            final String heightKey = defaultKey + ".default_height";
            final boolean useDefault = RuntimeGlobal.sysProp
                    .getBooleanWithDefault(defaultKey + ".default_use");
            final String defaultWidthStr;
            final String defaultHeightStr;
            if (useDefault) {
                defaultWidthStr = RuntimeGlobal.sysProp.getStringWithDefault(widthKey);
                defaultHeightStr = RuntimeGlobal.sysProp.getStringWithDefault(heightKey);
            } else if ("note".equals(kind) || "component".equals(kind)) {
                defaultWidthStr = RuntimeGlobal.sysProp.getDefaultString(widthKey);
                defaultHeightStr = RuntimeGlobal.sysProp.getDefaultString(heightKey);
            } else {
                return;
            }

            final Size size;
            try {
                final double defaultWidth = Double.parseDouble(defaultWidthStr);
                final double defaultHeight = Double.parseDouble(defaultHeightStr);
                size = NotationFactory.eINSTANCE.createSize(defaultWidth, defaultHeight);
            } catch (final NumberFormatException e) {
                // デフォルト値が不正な場合、sizeは設定しない
                LOGGER.warn("The value of the default size is not numeric. : " + widthKey + " or "
                        + heightKey);
                return;
            }
            append(SetCommand.create(domain, object, NotationPackage.Literals.SHAPE__SIZE, size));
        }
    }
}
