package net.astah.stpa.stamp.notation.edit.layout;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import net.astah.dd.dc.Bounds;
import net.astah.dd.dc.DCFactory;
import net.astah.dd.dc.Dimension;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.notation.Location;
import net.astah.notation.LocationConstraint;
import net.astah.notation.Shape;
import net.astah.notation.edit.layout.FreeLayout;
import net.astah.notation.edit.layout.OwnedElementsDelegatedLayout;
import net.astah.notation.edit.layout.TextAreaLayout;
import net.astah.notation.provider.GeometryProviderHelper;
import net.astah.stpa.stamp.notation.STAMPCompartment;
import net.astah.stpa.stamp.notation.STAMPCompartmentableShape;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.provider.STAMPCompartmentableShapeItemProvider;

/**
 * レイアウト処理で使用する共通クラスです。
 */
public class LayoutUtil {
    /**
     * {@link TextAreaLayout} の ID
     */
    public static final String TEXT_AREA = "text_area";

    /**
     * {@link FreeLayout} の ID
     */
    public static final String FREE = "free";

    /**
     * {@link OwnedElementsDelegatedLayout} の ID
     */
    public static final String COMPARTMENT = "compartment";

    /**
     * レイアウト処理クラスの ID と図要素のリストを対応させた {@link Map} を作成します。
     *
     * @param children
     *            {@link Map} 作成対象の図要素
     * @return レイアウト処理クラスの ID と図要素のリストを対応させた {@link Map} を作成します。
     */
    public static Map<String, List<DiagramElement>> createChildrenMap(
            final Iterable<? extends DiagramElement> children) {
        final Map<String, List<DiagramElement>> childrenMap = new HashMap<>();
        for (final DiagramElement child : children) {
            if (child instanceof STAMPLabel) {
                addChild(TEXT_AREA, child, childrenMap);
            } else if (child instanceof STAMPCompartmentableShape) {
                addChild(FREE, child, childrenMap);
            } else if (child instanceof STAMPCompartment) {
                addChild(COMPARTMENT, child, childrenMap);
            }
        }
        return childrenMap;
    }

    private static void addChild(final String id, final DiagramElement child,
            final Map<String, List<DiagramElement>> childrenMap) {
        List<DiagramElement> children = null;
        if (childrenMap.containsKey(id)) {
            children = childrenMap.get(id);
        } else {
            children = new ArrayList<DiagramElement>();
            childrenMap.put(id, children);
        }
        children.add(child);
    }

    /**
     * コンポーネント名の {@link Bounds} を取得します。
     * <p>
     * 子コンポーネントがある場合は、その影響も考慮されます。
     *
     * @param component
     *            コンポーネント
     * @param layoutCalculator
     *            使用する {@link LayoutCalculator}
     * @return コンポーネント名の {@link Bounds}
     */
    public static Bounds getComponentNameBounds(final Shape component,
            final LayoutCalculator layoutCalculator) {
        final TextAreaLayout textAreaLayout = (TextAreaLayout) STAMPCompartmentableShapeItemProvider.LAYOUT
                .getOwnedLayoutById(TEXT_AREA);
        final FreeLayout freeLayout = (FreeLayout) STAMPCompartmentableShapeItemProvider.LAYOUT
                .getOwnedLayoutById(FREE);
        final List<DiagramElement> elements = layoutCalculator.getOwnedElements(component);
        final Bounds totalBounds = DCFactory.eINSTANCE.createBounds();
        freeLayout.calculateTotalBounds(component, null, createChildrenMap(elements),
                layoutCalculator, totalBounds);
        double x = 0;
        double y = 0;
        double width = 0;
        if (totalBounds.getWidth() != 0) {
            width = totalBounds.getWidth();
            final LocationConstraint locationConstraint = GeometryProviderHelper
                    .getLocation(component, layoutCalculator);
            if (locationConstraint instanceof Location) {
                final Location location = (Location) locationConstraint;
                y = location.getY();
                if (location.getX() < totalBounds.getX()) {
                    width += totalBounds.getX() - location.getX();
                } else {
                    x = totalBounds.getX() - location.getX();
                }
            }
        }
        final Dimension size = DCFactory.eINSTANCE.createDimension(width, 0);
        textAreaLayout.calculateTotalSize(
                elements.stream().filter(e -> e instanceof STAMPLabel).collect(Collectors.toList()),
                layoutCalculator, size);
        return DCFactory.eINSTANCE.createBounds(x, y, size.getWidth(), size.getHeight());
    }
}
