package net.astah.stpa.stamp.notation.provider;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.astah.dd.dc.Point;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.provider.Geometries;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.golf.geom2D.Rectangle2d;
import net.astah.jmodel.RightAngleUtil;
import net.astah.notation.Edge;
import net.astah.notation.RightAngleLine;
import net.astah.notation.provider.GeometryProviderHelper;

/**
 * 直角線モードのとき垂直/水平線になるような制御点を返すレイアウト計算処理を実装したクラスです。
 */
public class STAMPGeometryProviderHelper extends GeometryProviderHelper {

    /**
     * 以下の条件が揃った時、ターゲットアンカーを基準とした垂直/水平線になるような{@code edge}の制御点を返す.<br>
     * <ul>
     * <li>直角線モードである</li>
     * <li>前回の {@code edge} が垂直/水平線である</li>
     * <li>ターゲットアンカーの座標とソース要素の矩形がX軸またはY軸で重なっている</li>
     * <li>直角線を引いたときの制御点の数が4以下である</li>
     * <li>X軸/Y軸で重なっているとき、線が縦/横から出入りしている</li>
     * </ul>
     * 垂直/水平線が引けない場合は
     * {@link GeometryProviderHelper#getControlPoints(Edge, Point, LayoutCalculator)}
     * に委譲する.
     * 
     * @param edge
     *            対象の{@code Edge}
     * @param origin
     *            原点の絶対座標
     * @param layoutCalculator
     *            使用する{@code LayoutCalculator}
     * @return {@code edge}の制御点
     */
    public static List<Point> getControlPoints(final Edge edge, final Point origin,
            final LayoutCalculator layoutCalculator) {
        if (canStraightLine(edge, origin, layoutCalculator)) {
            final DiagramElement target = layoutCalculator.getTarget(edge);
            final List<Point> controlPoints = getAlignedAnchorPoints(target, edge, layoutCalculator,
                    origin);
            if (!controlPoints.isEmpty()) {
                return controlPoints;
            }
        }
        return GeometryProviderHelper.getControlPoints(edge, origin, layoutCalculator);
    }

    /**
     * {@code edge}が垂直/水平線を引けるかを判定する.
     * 
     * @param edge
     *            対象の{@code Edge}
     * @param origin
     *            原点の絶対座標
     * @param layoutCalculator
     *            使用する{@code LayoutCalculator}
     * @return 垂直/水平線を引けるならtrue.
     */
    private static boolean canStraightLine(final Edge edge, final Point origin,
            final LayoutCalculator layoutCalculator) {
        if (!(getLine(edge, layoutCalculator) instanceof RightAngleLine)) {
            return false;
        }
        final DiagramElement source = layoutCalculator.getSource(edge);
        final DiagramElement target = layoutCalculator.getTarget(edge);
        if (source == target) {
            return false;
        }

        // ビューエレメントを表示していないとedge.wayPointに古い値が入っている場合があるが、今回は問題ないのでそのまま使用する。
        final List<Point> lastWayPoints = edge.getWaypoints();
        final List<Point> rightAngleControlPoints = GeometryProviderHelper.getControlPoints(edge,
                origin, layoutCalculator);
        final Rectangle2d sourceRect = Geometries
                .toRectangle2d(layoutCalculator.getSourceEndBounds(edge, origin));
        final Rectangle2d targetRect = Geometries
                .toRectangle2d(layoutCalculator.getTargetEndBounds(edge, origin));

        if (RightAngleUtil.isOverlapXCoordinate(sourceRect, targetRect)) {
            if (rightAngleControlPoints.size() <= 4
                    && RightAngleUtil.isOverlapXCoordinate(sourceRect,
                            Geometries.toPnt2d(rightAngleControlPoints.get(1)))
                    && RightAngleUtil.isOverlapXCoordinate(targetRect, Geometries.toPnt2d(
                            rightAngleControlPoints.get(rightAngleControlPoints.size() - 2)))
                    && RightAngleUtil.isOverlapXCoordinate(sourceRect, Geometries.toPnt2d(
                            rightAngleControlPoints.get(rightAngleControlPoints.size() - 1)))) {
                if (lastWayPoints.size() <= 2) {
                    return true;
                } else if (rightAngleControlPoints.size() == 4
                        && (Math.abs(rightAngleControlPoints.get(1).getX()
                                - rightAngleControlPoints.get(2).getX()) < 10)) {
                    return true;
                }
            }
        } else if (RightAngleUtil.isOverlapYCoordinate(sourceRect, targetRect)) {
            if (rightAngleControlPoints.size() <= 4
                    && RightAngleUtil.isOverlapYCoordinate(sourceRect,
                            Geometries.toPnt2d(rightAngleControlPoints.get(1)))
                    && RightAngleUtil.isOverlapYCoordinate(targetRect, Geometries.toPnt2d(
                            rightAngleControlPoints.get(rightAngleControlPoints.size() - 2)))
                    && RightAngleUtil.isOverlapYCoordinate(sourceRect, Geometries.toPnt2d(
                            rightAngleControlPoints.get(rightAngleControlPoints.size() - 1)))) {
                if (lastWayPoints.size() <= 2) {
                    return true;
                } else if (rightAngleControlPoints.size() == 4
                        && (Math.abs(rightAngleControlPoints.get(1).getY()
                                - rightAngleControlPoints.get(2).getY()) < 10)) {
                    return true;
                }
            }
        }

        return false;
    }

    private static List<Point> getAlignedAnchorPoints(final DiagramElement base, final Edge edge,
            final LayoutCalculator layoutCalculator, final Point origin) {
        final List<Point> controlPoints = new ArrayList<Point>();
        final DiagramElement source = layoutCalculator.getSource(edge);
        final DiagramElement target = layoutCalculator.getTarget(edge);
        controlPoints.add(layoutCalculator.getSourceEndPoint(edge, origin));
        controlPoints.add(layoutCalculator.getTargetEndPoint(edge, origin));

        final Point alignPoint;
        final Point basePoint;
        final Rectangle2d alignRect;
        if (base == source) {
            basePoint = controlPoints.get(0);
            alignPoint = controlPoints.get(1);
            alignRect = Geometries.toRectangle2d(layoutCalculator.getTargetEndBounds(edge, origin));
        } else if (base == target) {
            basePoint = controlPoints.get(1);
            alignPoint = controlPoints.get(0);
            alignRect = Geometries.toRectangle2d(layoutCalculator.getSourceEndBounds(edge, origin));
        } else {
            return Collections.emptyList();
        }

        if (basePoint == null) {
            return Collections.emptyList();
        }
        if (RightAngleUtil.isOverlapXCoordinate(alignRect, Geometries.toPnt2d(basePoint))) {
            alignPoint.setX(basePoint.getX());
        } else if (RightAngleUtil.isOverlapYCoordinate(alignRect, Geometries.toPnt2d(basePoint))) {
            alignPoint.setY(basePoint.getY());
        } else {
            return Collections.emptyList();
        }
        return controlPoints;
    }
}
