package net.astah.stpa.stamp.notation.editor;

import static java.util.Collections.emptyList;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;

import javax.swing.ImageIcon;

import org.eclipse.emf.common.notify.Notifier;

import net.astah.dd.dc.Bounds;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.golf.draw.GNode;
import net.astah.golf.draw.GRect;
import net.astah.golf.util.RuntimeGlobal;
import net.astah.golf.widget.CImageIcon;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ISelectionChangedListener;
import net.astah.golf.widget.SelectionChangedEvent;
import net.astah.golf.widget.swing.Selections;
import net.astah.golf.widget.swing.SwingImageIcon;
import net.astah.jomt.jview.Highlighter;
import net.astah.jomt.jview.IconResource;
import net.astah.notation.editor.DiagramViewElementImpl;
import net.astah.notation.editor.IViewElement;
import net.astah.notation.editor.ViewElements;
import net.astah.notation.editor.adjuster.AlignmentGuide;
import net.astah.notation.editor.adjuster.EvenGuide;
import net.astah.notation.editor.adjuster.EvenSizeGuide;
import net.astah.notation.editor.adjuster.FixedAspectRatioAdjuster;
import net.astah.notation.editor.adjuster.LinearAdjuster;
import net.astah.notation.editor.adjuster.MovingAdjuster;
import net.astah.notation.editor.adjuster.MovingCompoundAdjuster;
import net.astah.notation.editor.adjuster.OnGridAdjuster;
import net.astah.notation.editor.adjuster.ResizingAdjuster;
import net.astah.notation.editor.adjuster.ResizingCompoundAdjuster;
import net.astah.stpa.stamp.Element;
import net.astah.stpa.stamp.notation.STAMPDiagram;
import net.astah.stpa.stamp.notation.STAMPDiagramElement;

/**
 * 図を表示するためのビューエレメントを実装する抽象クラスです。
 */
public abstract class STAMPDiagramViewElement
        extends DiagramViewElementImpl<STAMPDiagram, STAMPDiagramElement>
        implements ISelectionChangedListener {
    private GRect dgmFrame = null;
    private boolean isCurrent;
    private boolean isFocus;

    private STAMPDiagramDrawSuggest drawSuggest;
    private Collection<Highlighter> currentHighlighters = new HashSet<>();

    private static final MovingAdjuster.Factory DEFAULT_MOVING_ADJUSTER = getDefaultMovingAdjusterFactory();
    private static final ResizingAdjuster.Factory DEFAULT_RESIZING_ADJUSTER = getDefaultResizingAdjusterFactory();

    public STAMPDiagramViewElement(final STAMPDiagram object) {
        super(object);
        updateDgmFrame();
        initDrawSuggest();
    }

    @Override
    public <T> T getAdapter(Class<T> adapter) {
        if (adapter.isInstance(drawSuggest)) {
            return adapter.cast(drawSuggest);
        } else if (adapter.isInstance(DEFAULT_MOVING_ADJUSTER)) {
            return adapter.cast(DEFAULT_MOVING_ADJUSTER);
        } else if (adapter.isInstance(DEFAULT_RESIZING_ADJUSTER)) {
            return adapter.cast(DEFAULT_RESIZING_ADJUSTER);
        }
        return super.getAdapter(adapter);
    }

    private void initDrawSuggest() {
        drawSuggest = new STAMPDiagramDrawSuggest(this);
        CImageIcon image = new SwingImageIcon((ImageIcon) IconResource.getNormalSizeAssocIcon());
        drawSuggest.addCommandAction("net.astah.jomt.jcontrol.CreateShape", 2, image);
    }

    private void updateDgmFrame() {
        dgmFrame = new GRect(dgmRect);
        dgmFrame.setColorNo((byte) 6);
        dgmFrame.setFilled(false); // for performance
        dgmFrame.setDepth(Integer.MAX_VALUE);
        dgmFrame.setSelectable(false);

        if (RuntimeGlobal.sysProp != null) {
            boolean isDgmFrameVisibility = RuntimeGlobal.sysProp
                    .getBooleanWithDefault("basic.diagram_frame_visibility");
            dgmFrame.setVisible(isDgmFrameVisibility);
        }
    }

    @Override
    protected Bounds getTotalBounds(List<DiagramElement> elements) {
        return new LayoutCalculator().getTotalBounds(elements, true);
    }

    private static MovingAdjuster.Factory getDefaultMovingAdjusterFactory() {
        return new MovingAdjuster.Factory() {
            @Override
            public MovingAdjuster createAdjuster() {
                return new MovingCompoundAdjuster().add(new LinearAdjuster())
                        .add(new OnGridAdjuster())
                        .add(new AlignmentGuide()).add(new EvenGuide());
            }
        };
    }

    private static ResizingAdjuster.Factory getDefaultResizingAdjusterFactory() {
        return new ResizingAdjuster.Factory() {
            @Override
            public ResizingAdjuster createAdjuster() {
                return new ResizingCompoundAdjuster().add(new FixedAspectRatioAdjuster())
                        .add(new EvenSizeGuide());
            }
        };
    }

    @Override
    public void update(Notifier notifier, Object object) {
    }

    @Override
    public boolean isDiagramFrameReady() {
        return (dgmFrame != null);
    }

    @Override
    public double getX() {
        return dgmFrame.getX();
    }

    @Override
    public double getY() {
        return dgmFrame.getY();
    }

    @Override
    public double getWidth() {
        return dgmFrame.getWidth();
    }

    @Override
    public double getHeight() {
        return dgmFrame.getHeight();
    }

    @Override
    public boolean isCurrent() {
        return isCurrent;
    }

    @Override
    public void setCurrent(boolean isCurrent) {
        this.isCurrent = isCurrent;
    }

    @Override
    public boolean isFocus() {
        return isFocus;
    }

    @Override
    public void setFocus(boolean isFocus) {
        this.isFocus = isFocus;
    }

    @Override
    public GNode getBodyNode() {
        return dgmFrame;
    }

    @Override
    protected List<? extends STAMPDiagramElement> getOwnedElements() {
        final STAMPDiagram model = getModel();
        if (model != null) {
            return model.getSTAMPOwnedElements();
        }
        return emptyList();
    }

    @Override
    public void setCreateEditState(boolean isEditState) {
    }

    @Override
    public boolean getCreateEditState() {
        return false;
    }

    @Override
    public void layout(final LayoutCalculator layoutCalculator) {
        super.layout(layoutCalculator);
        updateDgmFrame();
    }

    @Override
    public void selectionChanged(SelectionChangedEvent event) {
        Collection<Highlighter> highlighters = getHighlighters(event.getSelection());
        currentHighlighters.forEach(h -> {
            if (!highlighters.contains(h)) {
                h.setHighlighted(false);
            }
        });
        highlighters.forEach(h -> {
            if (!currentHighlighters.contains(h)) {
                h.setHighlighted(true);
            }
        });
        currentHighlighters.clear();
        currentHighlighters.addAll(highlighters);
    }

    private Collection<Highlighter> getHighlighters(final ISelection selection) {
        if (Selections.size(selection) != 1) {
            return Collections.emptyList();
        }
        Object target = Selections.getFirstElement(selection);
        if (target instanceof IViewElement) {
            Highlighter highlighter = ((IViewElement) target).getAdapter(Highlighter.class);
            if (highlighter != null) {
                return highlighter.getHighlightTogether();
            }
        } else if (target instanceof Element && isHighlightModel((Element) target)) {
            for (GNode child : getAllNodes()) {
                if (child.getModel() instanceof DiagramElement) {
                    DiagramElement diagramElement = (DiagramElement) child.getModel();
                    if (diagramElement.getModelElements().contains(target)) {
                        IViewElement ve = ViewElements.getAncestor(child, IViewElement.class);
                        Highlighter highlighter = ve.getAdapter(Highlighter.class);
                        if (highlighter != null) {
                            return Arrays.asList(highlighter);
                        }
                    }
                }
            }
        }
        return Collections.emptyList();
    }

    /**
     * 与えられた {@code Element} がハイライト対象かどうかを返します。
     * 
     * @param target
     *            判定エレメント
     * @return true：ハイライト対象である
     */
    protected boolean isHighlightModel(final Element target) {
        return true;
    }
}
