package net.astah.stpa.stamp.ui.control;

import java.util.Collections;
import java.util.stream.Collectors;

import org.apache.commons.lang3.ArrayUtils;

import com.google.common.collect.Iterables;

import net.astah.common.collect.Streams;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.ecore.EComparators;
import net.astah.emf.common.notify.EFeaturePath;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.DiagramEditor;
import net.astah.jomt.jview.MessageManager;
import net.astah.stpa.stamp.ControlLink;
import net.astah.stpa.stamp.STAMPPackage;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPNotationPackage;
import net.astah.stpa.stamp.notation.STAMPShape;
import net.astah.stpa.stamp.ui.Messages;

/**
 * コントロールループ図を作成するコマンドです。
 * <P>
 * 主に、マネジメントビューなどのメニューからのアクションイベントによる呼び出しを想定していてます。
 * </P>
 */
public class CreateControlLoopDiagramFromPrjCommand extends CreateControlLoopDiagramCommand {

    /**
     * 引数に指定した場合、既存のコントロールループ図があれば開く
     * 
     * @see #setArgumentString(String)
     */
    public final static String OPEN_IF_EXISTS = "openIfExists";

    private boolean openIfExists = false;

    public CreateControlLoopDiagramFromPrjCommand() {
        super();
        setSelectedDiagramElement();
    }

    private void setSelectedDiagramElement() {
        DiagramEditor currentDiagramEditor = JomtGlobal.viewManager.getCurrentDiagramEditor();
        if (currentDiagramEditor == null) {
            return;
        }

        Object[] selectedModels = currentDiagramEditor.getSelectedModels();
        if (ArrayUtils.isEmpty(selectedModels)) {
            return;
        }

        DiagramElement selected = (DiagramElement) selectedModels[0];
        if (selected instanceof STAMPShape && isControlLink(selected.getOwningElement())) {
            selected = selected.getOwningElement();
        }

        setSelectedDiagramElement(selected);
    }

    @Override
    public void setArgumentString(final String arg) {
        openIfExists = OPEN_IF_EXISTS.equalsIgnoreCase(arg);
    }

    @Override
    public void execute() {
        if (openIfExists) {
            final Iterable<STAMPControlLoopDiagram> clDiagrams = getCLDiagrams(
                    getSelectedDiagramElement());
            if (!Iterables.isEmpty(clDiagrams)) {
                final STAMPControlLoopDiagram diagram = selectCLDiagram(clDiagrams);
                if (diagram != null) {
                    callCommand(createOpenEditorCommand(diagram));
                }
                return;
            }
        }

        if (!isControlLink(getSelectedDiagramElement())) {
            MessageManager.displayErrorMessageDialog(Messages.cl_cannot_create);
            return;
        }

        super.execute();
    }

    private boolean isControlLink(Object object) {
        return object instanceof STAMPEdge
                && DiagramElements.isModelElementPresent((STAMPEdge) object, ControlLink.class);
    }

    private Iterable<STAMPControlLoopDiagram> getCLDiagrams(final DiagramElement diagramElement) {
        if (diagramElement == null) {
            // TODO 将来的に、既存のCL図を選択して開けるようにするなら、ここで既存全CL図を返すようにする
            return Collections.emptyList();
        }
        final ControlLink controlLink = DiagramElements.getModelElement(diagramElement,
                ControlLink.class);
        if (controlLink == null) {
            // TODO 将来的に、既存のCL図を選択して開けるようにするなら、ここで既存全CL図を返すようにする
            return Collections.emptyList();
        }
        final EFeaturePath path = EFeaturePath.builder()
                .addInverse(STAMPPackage.Literals.CONTROL_LOOP__TARGET_LINK)
                .addInverse(STAMPNotationPackage.Literals.STAMP_CONTROL_LOOP_DIAGRAM__CONTROL_LOOP)
                .create();
        return Streams.stream(path.getReferences(controlLink), STAMPControlLoopDiagram.class)
                .sorted(EComparators.eObjectPath()).collect(Collectors.toList());
    }

    private STAMPControlLoopDiagram selectCLDiagram(
            final Iterable<STAMPControlLoopDiagram> diagrams) {
        // TODO 将来的に、既存のCL図を選択して開けるようにするなら、ここで選択ダイアログを表示する
        return Iterables.getFirst(diagrams, null);
    }

}
