package net.astah.stpa.stamp.ui.control.mode;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.List;

import org.eclipse.emf.ecore.EModelElement;

import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.command.DiagramContext;
import net.astah.dd.di.edit.provider.Geometries;
import net.astah.dd.di.util.DiagramElements;
import net.astah.golf.control.CommandEvent;
import net.astah.golf.draw.GNode;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.golf.view.swing.InputEventSupport;
import net.astah.jomt.jcontrol.mode.DiagramMode;
import net.astah.jomt.jcontrol.mode.DrawingRubberSupport;
import net.astah.jomt.jviewElement.JomtNode;
import net.astah.notation.editor.IViewElement;
import net.astah.notation.editor.ViewElements;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.STAMPShape;
import net.astah.ui.control.CreateShapeCommand;

/**
 * コンポーネントを作成する Mode です。
 */
public class CreateComponentMode extends DiagramMode {

    protected STAMPComponentShape owingElement = null;

    protected GNode containerRect = null;

    public CreateComponentMode() {
        containerRect = new GNode();
    }

    @Override
    public void mouseClicked(MouseEvent e) {
        e.consume();
    }

    /**
     * Shape のインスタンスを返します。
     * 
     * @return Shape を作成してインスタンス
     */
    protected STAMPShape createShape() {
        return STAMPNotationFactory.eINSTANCE.createSTAMPComponentShape();
    }

    /**
     * モデルのインスタンスを作成して返します。
     * 
     * @return モデルのインスタンス
     */
    protected EModelElement createModelElement() {
        return STAMPFactory.eINSTANCE.createComponent();
    }

    /**
     * Shape のインスタンスを作成し、初期化してから返します。
     * 
     * @return Shape のインスタンス
     */
    protected STAMPShape prepareShape() {
        final STAMPShape shape = createShape();
        // 図要素の登録に先立ってモデル要素を追加しておく必要がある
        DiagramElements.setModelElement(shape, createModelElement());
        return shape;
    }

    /**
     * Shape を作成するコマンドを作成して返します。
     * 
     * @return Shape を作成するコマンド
     */
    protected CreateShapeCommand createCommand() {
        return new CreateShapeCommand();
    }

    /**
     * コンポーネント上でマウスボタンが離されたときに呼び出されます。
     */
    @Override
    public void mouseReleased(MouseEvent e) {
        if (!InputEventSupport.isLeftMouseButton(e)) {
            return;
        }

        double x = camera.toWorldCoordX(e.getX());
        double y = camera.toWorldCoordY(e.getY());

        Pnt2d p = new Pnt2d(x, y);

        CreateShapeCommand command = createCommand();
        command.setShape(prepareShape());

        command.setLocation(Geometries.toPoint(p));
        command.setContext(new DiagramContext(getDiagram(), null, p, owingElement));
        DiagramElement owner = (owingElement == null) ? getDiagram() : owingElement;
        command.setOwningElement(owner);
        command.setShiftPressedState(e.isShiftDown());

        fireCommandPerformed(new CommandEvent(e.getSource(), e.getID(), "CreateComponent",
                command, e.getModifiers()));
        clear();
        e.consume();
    }

    /**
     * コンポーネント上でマウスボタンが押されたときに呼び出されます。
     */
    @Override
    public void mousePressed(MouseEvent e) {
        if (InputEventSupport.isRightMouseButton(e)) {
            return;
        }
        e.consume();
    }

    /**
     * コンポーネント上でマウスボタンが押されてからドラッグされたときに呼び出されます。 （ マウスの位置がコンポーネントの範囲内にあるかどうかにかかわらず ）
     * マウスボタンが離されるまで発生します。
     */
    @Override
    public void mouseDragged(MouseEvent e) {
        if (InputEventSupport.isRightMouseButton(e)) {
            return;
        }
        mouseMoved(e);
    }

    /**
     * コンポーネント上でマウスボタンが移動したときに呼び出されます。 （ ボタンが押されていない状態 ）
     */
    @Override
    public void mouseMoved(MouseEvent e) {
        // ラバー表示の初期化
        DrawingRubberSupport.clearOldRubber(canvas);
        scene.removeTempNode(containerRect);
        containerRect.removeAllChildren();

        IViewElement owningViewElement = getOwningViewElement(getMouseLocation(e));
        if (owningViewElement != null
                && owningViewElement.getModel() instanceof STAMPComponentShape) {
            containerRect.addChild(owningViewElement.createFeedbackShape());
            owingElement = (STAMPComponentShape) owningViewElement.getModel();
            canvas.repaintSync();
            canvas.drawTempFront();
        } else {
            owingElement = null;
        }
        if (!containerRect.getChildren().isEmpty()) {
            scene.addTempNode(containerRect);
        }

        DrawingRubberSupport.drawNewRubber(canvas);

        e.consume();
    }

    @Override
    public void keyReleased(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
            clear();
        }
    }

    private void clear() {
        scene.removeTempNode(containerRect);
        canvas.clear();
        canvas.draw();
        canvas.repaint();
        containerRect.removeAllChildren();
    }

    private IViewElement getOwningViewElement(Pnt2d mouseLocation) {
        final List<GNode> pickedNodes = ignoreRimmedPick(() -> {
            return getPickedNodes(mouseLocation, JomtNode.BODY, false);
        });
        pickedNodes.removeIf(node -> !(node.getModel() instanceof STAMPComponentShape));
        if (!pickedNodes.isEmpty()) {
            GNode minDepthNode = getMinDepthNode(pickedNodes);
            return ViewElements.getAncestor(minDepthNode, IViewElement.class);
        }
        return null;
    }
}
