package net.astah.stpa.stamp.ui.menu.handler;

import static java.util.Collections.singleton;
import static net.astah.stpa.stamp.STAMPPackage.Literals.ACCIDENT__HAZARD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.STPA_ANALYSIS__HAZARD;
import static net.astah.table.TablePackage.Literals.ELEMENT__OBJECT;
import static net.astah.table.TablePackage.Literals.GRID__ROW_TUPLE;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.awt.event.ActionEvent;
import java.util.List;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.golf.widget.IListSelection;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ITableSelection;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.Editor;
import net.astah.jomt.jview.EditorPeer;
import net.astah.stpa.stamp.Accident;
import net.astah.stpa.stamp.Hazard;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.ui.view.swing.AccidentHazardSafetyConstraintTableEditorPeer;
import net.astah.table.Element;
import net.astah.table.Grid;
import net.astah.table.Tuple;
import net.astah.table.edit.command.AddRowCommand;
import net.astah.table.ui.TableEditors;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * アクシデント表にハザードを追加します。
 */
@Handler
@MenuItem(commands = @MenuCommand(id = "stpa.AddHazardCommand",
        category = "edit.table"),
        label = "%StampTable_addHazard_label",
        tooltip = "%StampTable_addHazard_tooltip",
        type = ItemKind.PUSH)
public class AddHazardHandler {

    private final int HAZARD_ID_COLUMN = 2;
    private final int HAZARD_COLUMN = 3;

    private Command command = UnexecutableCommand.INSTANCE;

    private int newRow = -1;

    /**
     * アクシデント表にハザードを追加するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final CommandJob job = new CommandJob(getDomain(), command);
        Jobs.toCompletableFuture(job, true).thenRun(() -> {
            TableEditors.startCellEditing(newRow, HAZARD_COLUMN);
        });
        job.schedule();
    }

    private TransactionalEditingDomain getDomain() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        return TransactionUtil.getEditingDomain(editor.getEditModel());
    }

    /**
     * アクシデント表にハザードを追加するコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        command.dispose();
        command = UnexecutableCommand.INSTANCE;
        if (selection instanceof ITableSelection) {
            command = getCommand((ITableSelection) selection);
        }
        return command.canExecute();
    }

    private Command getCommand(final ITableSelection selection) {
        final Object element = selection.getRowSelection().getFirstElement();
        if (element instanceof Tuple) {
            if (isHazard(selection.getColumnSelection())) {
                return createCommand((Tuple) element);
            }
        }
        return command;
    }

    private boolean isHazard(final IListSelection columnSelection) {
        final int[] indices = columnSelection.getIndices();
        if (indices.length > 0) {
            return indices[0] == HAZARD_ID_COLUMN
                    || indices[0] == HAZARD_COLUMN;
        }
        return false;
    }

    private Command createCommand(final Tuple tuple) {
        final int row = tuple.getIndex();
        final Grid grid = tuple.getGrid();
        final EList<Tuple> rowTuples = grid.getRowTuples();
        if ((row >= 0) && (row < rowTuples.size())) {
            final Tuple rowTuple = rowTuples.get(row);
            return createAddHazardCommand(grid, row, rowTuple);
        }
        return command;
    }

    private Command createAddHazardCommand(final Grid grid, final int row, final Tuple rowTuple) {
        final AccidentHazardSafetyConstraintTableEditorPeer editorPeer = getEditorPeer();
        final STPAAnalysis scope = editorPeer.getScope();
        final EditingDomain domain = getEditingDomainFor(scope);
        final Hazard hazard = STAMPFactory.eINSTANCE.createHazard();
        final EList<Element> elements = rowTuple.getElements();
        final EObject accident = elements.get(editorPeer.ACCIDENT_INDEX).getObject();

        final Command command = CreateChildCommand.create(domain, scope,
                new CommandParameter(scope, STPA_ANALYSIS__HAZARD, hazard), singleton(scope))
                .chain(AddCommand.create(domain, accident, ACCIDENT__HAZARD, hazard));

        final Element element = elements.get(editorPeer.HAZARD_INDEX);
        if (element.getObject() == null) {
            // 現在の行にハザードが設定されていないので、現在行にハザードを追加する
            this.newRow = row;
            return command.chain(SetCommand.create(getDomain(), element, ELEMENT__OBJECT, hazard));
        } else {
            // 現在の行にハザードが設定されているので、新規行としてハザードを追加する
            EList<Hazard> hazards = ((Accident) accident).getHazards();
            int index = hazards.indexOf(element.getObject());
            this.newRow = row - index + hazards.size();
            final Tuple newRowTuple = AddRowCommand.prepareNewTuple(grid, row);
            final EList<Element> newElements = newRowTuple.getElements();
            newElements.get(editorPeer.ACCIDENT_INDEX).setObject(accident);
            newElements.get(editorPeer.HAZARD_INDEX).setObject(hazard);
            return AddCommand.create(getDomain(), grid, GRID__ROW_TUPLE, newRowTuple, row + 1)
                    .chain(command);
        }
    }

    private AccidentHazardSafetyConstraintTableEditorPeer getEditorPeer() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        final EditorPeer editorPeer = editor.getEditorPeer();
        return (AccidentHazardSafetyConstraintTableEditorPeer) editorPeer;
    }

}
