package net.astah.stpa.stamp.ui.menu.handler;

import static java.util.Collections.singleton;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD__SAFETY_CONSTRAINT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.STPA_ANALYSIS__SAFETY_CONSTRAINT;
import static net.astah.table.TablePackage.Literals.ELEMENT__OBJECT;
import static net.astah.table.TablePackage.Literals.GRID__ROW_TUPLE;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.awt.event.ActionEvent;
import java.util.List;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.golf.widget.IListSelection;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ITableSelection;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.Editor;
import net.astah.jomt.jview.EditorPeer;
import net.astah.stpa.stamp.Hazard;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.SafetyConstraint;
import net.astah.stpa.stamp.ui.view.swing.AccidentHazardSafetyConstraintTableEditorPeer;
import net.astah.table.Element;
import net.astah.table.Grid;
import net.astah.table.Tuple;
import net.astah.table.edit.command.AddRowCommand;
import net.astah.table.ui.TableEditors;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * アクシデント表に安全制約を追加します。
 */
@Handler
@MenuItem(commands = @MenuCommand(id = "stpa.AddSafetyConstraintCommand",
        category = "edit.table"),
        label = "%StampTable_addSafetyConstraint_label",
        tooltip = "%StampTable_addSafetyConstraint_tooltip",
        type = ItemKind.PUSH)
public class AddSafetyConstraintHandler {

    private final int SAFETY_CONSTRAINT_ID_COLUMN = 4;
    private final int SAFETY_CONSTRAINT_COLUMN = 5;

    private Command command = UnexecutableCommand.INSTANCE;

    private int newRow = -1;

    /**
     * アクシデント表に安全制約を追加するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final CommandJob job = new CommandJob(getDomain(), command);
        Jobs.toCompletableFuture(job, true).thenRun(() -> {
            TableEditors.startCellEditing(newRow, SAFETY_CONSTRAINT_COLUMN);
        });
        job.schedule();
    }

    private TransactionalEditingDomain getDomain() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        return TransactionUtil.getEditingDomain(editor.getEditModel());
    }

    /**
     * アクシデント表に安全制約を追加するコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        command.dispose();
        command = UnexecutableCommand.INSTANCE;
        if (selection instanceof ITableSelection) {
            command = getCommand((ITableSelection) selection);
        }
        return command.canExecute();
    }

    private Command getCommand(final ITableSelection selection) {
        final Object element = selection.getRowSelection().getFirstElement();
        if (element instanceof Tuple) {
            if (isSafetyConstraint(selection.getColumnSelection())) {
                final Tuple tuple = (Tuple) element;
                return createAddSafetyConstraintCommand(tuple.getGrid(), tuple.getIndex());
            }
        }
        return command;
    }

    private boolean isSafetyConstraint(final IListSelection columnSelection) {
        final int[] indices = columnSelection.getIndices();
        if (indices.length > 0) {
            return indices[0] >= SAFETY_CONSTRAINT_ID_COLUMN;
        }
        return false;
    }

    /**
     * 選択されているGridと行を元に安全制約を追加するコマンドを構築します。
     * 
     * @param grid
     *            選択されているGrid
     * @param row
     *            行
     * @return 構築したコマンド
     */
    public Command createAddSafetyConstraintCommand(final Grid grid, final int row) {
        final EList<Tuple> rowTuples = grid.getRowTuples();
        if ((row >= 0) && (row < rowTuples.size())) {
            final Tuple rowTuple = rowTuples.get(row);
            return createAddSafetyConstraintCommand(grid, row, rowTuple);
        }
        return command;
    }

    private Command createAddSafetyConstraintCommand(final Grid grid,
            final int row, final Tuple rowTuple) {
        final AccidentHazardSafetyConstraintTableEditorPeer editorPeer = getEditorPeer();
        final STPAAnalysis scope = editorPeer.getScope();
        final EditingDomain domain = getEditingDomainFor(scope);
        final SafetyConstraint safetyConstraint = STAMPFactory.eINSTANCE.createSafetyConstraint();
        final EList<Element> elements = rowTuple.getElements();
        final EObject accident = elements.get(editorPeer.ACCIDENT_INDEX).getObject();
        final Hazard hazard = (Hazard) elements.get(editorPeer.HAZARD_INDEX).getObject();

        final Command command = CreateChildCommand.create(domain, scope,
                new CommandParameter(scope, STPA_ANALYSIS__SAFETY_CONSTRAINT, safetyConstraint),
                singleton(scope))
                .chain(AddCommand.create(domain, hazard, HAZARD__SAFETY_CONSTRAINT,
                        safetyConstraint));

        final Element element = elements.get(editorPeer.SAFETY_CONSTRAINT_INDEX);
        if (element.getObject() == null) {
            // 現在の行に安全制約が設定されていないので、現在行に安全制約を追加する
            this.newRow = row;
            return command.chain(
                    SetCommand.create(getDomain(), element, ELEMENT__OBJECT, safetyConstraint));
        } else {
            // 現在の行に安全制約が設定されているので、新規行として安全制約を追加する
            this.newRow = row + 1;
            EList<SafetyConstraint> safetyConstraints = hazard.getSafetyConstraints();
            int index = safetyConstraints.indexOf(element.getObject());
            this.newRow = row - index + safetyConstraints.size();

            final Tuple newRowTuple = AddRowCommand.prepareNewTuple(grid, row);
            final EList<Element> newElements = newRowTuple.getElements();
            newElements.get(editorPeer.ACCIDENT_INDEX).setObject(accident);
            newElements.get(editorPeer.HAZARD_INDEX).setObject(hazard);
            newElements.get(editorPeer.SAFETY_CONSTRAINT_INDEX).setObject(safetyConstraint);
            return AddCommand.create(getDomain(), grid, GRID__ROW_TUPLE, newRowTuple, row + 1)
                    .chain(command);
        }
    }

    private AccidentHazardSafetyConstraintTableEditorPeer getEditorPeer() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        final EditorPeer editorPeer = editor.getEditorPeer();
        return (AccidentHazardSafetyConstraintTableEditorPeer) editorPeer;
    }
}
