package net.astah.stpa.stamp.ui.menu.handler;

import java.awt.event.ActionEvent;
import java.util.List;

import javax.swing.JMenuItem;

import org.eclipse.emf.common.util.EList;

import net.astah.dd.dg.Ellipse;
import net.astah.dd.dg.GraphicalElement;
import net.astah.dd.dg.Line;
import net.astah.dd.dg.Polyline;
import net.astah.dd.dg.Rectangle;
import net.astah.dd.dg.util.DGStyleResolver;
import net.astah.golf.control.GolfAction;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.jmodel.StyleSupport;
import net.astah.ui.menu.annotation.AboutToShow;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 図要素に線種を設定します。
 * 
 * <p>
 * 実線、細かい点線、点線。
 */
@Handler(commands = "stpa.SetLineTypeCommand")
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineTypeCommand",
        category = "edit.style"),
        path = "LineType",
        label = "%StampControlStructureDiagram_lineType_label",
        tooltip = "%StampControlStructureDiagram_lineType_tooltip",
        type = ItemKind.MENU)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineTypeCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_TYPE_SOLID),
        path = "LineType." + StyleSupport.STYLE_LINE_TYPE_SOLID,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineStyle1.png",
        type = ItemKind.CHECK)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineTypeCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_TYPE_DASH1),
        path = "LineType." + StyleSupport.STYLE_LINE_TYPE_DASH1,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineStyle2.png",
        type = ItemKind.CHECK)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineTypeCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_TYPE_DASH2),
        path = "LineType." + StyleSupport.STYLE_LINE_TYPE_DASH2,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineStyle3.png",
        type = ItemKind.CHECK)
public class SetLineTypeHandler {

    /**
     * 図要素に線種を設定するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final GolfAction action = new GolfAction(
                "net.astah.jomt.jcontrol.SetLineType%" + params.get(0));
        action.actionPerformed(e);
    }

    /**
     * 図要素の線種を設定するコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        // 選択要素が共通図要素のみの場合、true
        return !Selections.isEmpty(selection)
                && Selections.stream(selection).allMatch(object -> isCommonElement(object));
    }

    private boolean isCommonElement(final Object object) {
        if (object instanceof Line
                || object instanceof Rectangle
                || object instanceof Ellipse
                || object instanceof Polyline) {
            return true;
        }
        return false;
    }

    /**
     * メニュー表示の直前に、図要素の線種に該当するメニューをチェックします。
     * 
     * @param selection
     *            選択要素
     * @param menuItem
     *            表示するメニューアイテム
     * @param params
     *            コマンドに指定されているパラメータ
     */
    @AboutToShow
    public void aboutToShow(final ISelection selection, final JMenuItem menuItem,
            final List<String> params) {
        if (params.isEmpty()) {
            return;
        }
        if (Selections.size(selection) > 1) {
            return;
        }
        final boolean isSameLineType = isSameLineType(Selections.getFirstElement(selection),
                params.get(0));
        menuItem.setSelected(isSameLineType);
    }

    private boolean isSameLineType(final Object object, final String type) {
        if (object instanceof GraphicalElement) {
            final GraphicalElement element = (GraphicalElement) object;
            final String lineType = getLineType(element);
            return lineType.equals(type);
        }
        return false;
    }

    private String getLineType(final GraphicalElement element) {
        final EList<Double> dashLengths = DGStyleResolver.getInstance(element)
                .resolveStrokeDashLengths(element);
        String lineType = StyleSupport.STYLE_LINE_TYPE_SOLID;
        if (!dashLengths.isEmpty() && dashLengths.size() == 2) {
            final double length0 = dashLengths.get(0);
            final double length1 = dashLengths.get(1);
            if (length0 == length1) {
                lineType = StyleSupport.STYLE_LINE_TYPE_DASH1;
            } else if (length0 == (length1 * 2)) {
                lineType = StyleSupport.STYLE_LINE_TYPE_DASH2;
            }
        }
        return lineType;
    }

}