package net.astah.stpa.stamp.ui.menu.handler;

import java.util.List;
import java.util.MissingResourceException;

import javax.swing.Icon;
import javax.swing.JMenuItem;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.provider.ExtraLabelVariant;
import net.astah.emf.edit.provider.LabelProviders;
import net.astah.golf.util.UIFactory;
import net.astah.golf.widget.CIcon;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.golf.widget.swing.SwingIcon;
import net.astah.golf.widget.swing.SwingIconRegistry;
import net.astah.stpa.stamp.notation.STAMPCompartmentableShape;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.notation.STAMPControlStructureDiagram;
import net.astah.stpa.stamp.notation.STAMPDiagram;
import net.astah.ui.menu.annotation.AboutToShow;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 図要素が構造ツリーに表示されている場合、図要素から構造ツリー上のモデルへジャンプします。
 */
@Handler(targets = { STAMPCompartmentableShape.class,
        STAMPControlStructureDiagram.class,
        STAMPControlLoopDiagram.class })
@MenuItem(label = "%StampControlStructureDiagram_showInStructureTree_label",
        tooltip = "%StampControlStructureDiagram_showInStructureTree_tooltip",
        type = ItemKind.MENU)
@MenuCommand(id = "stpa.ShowInStructureTreeCommand",
        category = "edit.select_show")
public class ShowInStructureTreeHandler {

    private static final Logger LOG = LoggerFactory.getLogger(ShowInStructureTreeHandler.class);

    /**
     * 図要素から構造ツリー上のモデルへジャンプするコマンド実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        // 選択要素が1つ、かつ、構造ツリーに表示する要素の場合、trueとする。
        return Selections.size(selection) == 1;
    }

    /**
     * メニュー表示の直前に、構造ツリー上のモデルへジャンプするメニューに対して、サブメニューを構築します。
     * 
     * <p>
     * サブメニューには、モデルの名前、構造ツリー上のモデルへジャンプするアクション、モデルのアイコンを設定します。
     * 
     * @param selection
     *            選択要素
     * @param menuItem
     *            表示するメニューアイテム
     * @param params
     *            コマンドに指定されているパラメータ
     */
    @AboutToShow
    public void aboutToShow(final ISelection selection, final JMenuItem menuItem,
            final List<String> params) {
        if (Selections.size(selection) > 1) {
            return;
        }
        final Object object = Selections.getFirstElement(selection);
        if (!(object instanceof DiagramElement)) {
            return;
        }
        final EObject element;
        if (object instanceof STAMPDiagram) {
            element = (STAMPDiagram) object;
        } else {
            element = DiagramElements.getModelElement((DiagramElement) object);
        }
        if (element == null) {
            return;
        }
        final String name = LabelProviders.getText(element, ExtraLabelVariant.MODEL_TREE);
        final String action = "ShowInStructTree" + "%" + EcoreUtil.getURI(element);
        final CIcon cIcon = getIcon(element);
        menuItem.add(UIFactory.createMenuItem(name, action, cIcon));
    }

    private CIcon getIcon(final EObject element) {
        try {
            final Object image = LabelProviders.getImage(element);
            if (image != null) {
                final Icon icon = SwingIconRegistry.INSTANCE.getIcon(image);
                return new SwingIcon(icon);
            }
        } catch (final MissingResourceException e) {
            LOG.warn("Failed to get an image: {}", e.getMessage());
        }
        return null;
    }

}
