package net.astah.stpa.stamp.resource;

import static net.astah.stpa.stamp.STAMPPackage.Literals.HCF_HINT_WORD_SET;
import static net.astah.stpa.stamp.resource.STAMPResource.FILE_EXTENSION;
import static net.astah.stpa.stamp.resource.STAMPResource.HCF_HINT_WORD_SETS_LIBRARY_URI;
import static net.astah.stpa.stamp.resource.STAMPResource.LIBRARIES_PATHMAP;

import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.URIConverter;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;

import net.astah.stpa.stamp.HCFHintWordSet;
import net.astah.stpa.stamp.STAMPPlugin;
import net.astah.stpa.stamp.STPAAnalysis;

/**
 * STAMP/STPA モデルに関するリソースを扱うためのユーティリティクラスです。
 */
public class STAMPResources {

    /**
     * 指定された {@code ResourceSet} を STAMP/STPA メタモデルを扱えるように構成します。
     * 
     * @param resourceSet
     *            構成する {@code ResourceSet}
     */
    public static void init(final ResourceSet resourceSet) {
        initLocalRegistries(resourceSet);
        initGlobalRegistries();
    }

    /**
     * 指定された {@code ResourceSet} を STAMP/STPA メタモデルを扱えるように構成します。
     * <p>
     * このメソッドでは指定された {@code resourceSet} に対する設定のみを行います。
     * </p>
     * 
     * @param resourceSet
     *            構成する {@code ResourceSet}
     */
    public static void initLocalRegistries(final ResourceSet resourceSet) {
        initURIConverterURIMap(resourceSet.getURIConverter().getURIMap());
        initResourceFactoryRegistry(resourceSet.getResourceFactoryRegistry());
    }

    /**
     * グローバルなレジストリに対して STAMP/STPA メタモデルを扱うための登録を行います。
     */
    public static void initGlobalRegistries() {
        initURIConverterURIMap(URIConverter.URI_MAP);
        initResourceFactoryRegistry(Resource.Factory.Registry.INSTANCE);
    }

    /**
     * {@code URI} の読み替えを行うマッピングを登録します。
     * 
     * @param uriMap
     *            マッピングを登録する先の {@code Map}
     * @return {@code uriMap} で指定されたインスタンス
     */
    public static Map<URI, URI> initURIConverterURIMap(final Map<URI, URI> uriMap) {
        final URI baseURI = URI.createURI(STAMPPlugin.INSTANCE.getBaseURL().toString());
        uriMap.put(URI.createURI(LIBRARIES_PATHMAP),
                URI.createURI("model/libraries/").resolve(baseURI));
        return uriMap;
    }

    /**
     * STAMP/STPA 用の {@code Resource.Factory} をレジストリに登録します。
     * 
     * @param registry
     *            登録先のレジストリ
     */
    public static void initResourceFactoryRegistry(final Resource.Factory.Registry registry) {
        registry.getExtensionToFactoryMap().put(FILE_EXTENSION, STAMPResource.Factory.INSTANCE);
    }

    /**
     * システム組み込みのヒントワードセットのリソースから {@code fragment} を持つヒントワードセットの {@code URI} を返します。
     * 
     * @param fragment
     *            取得するヒントワードセットの ID を示す URI フラグメント
     * @return {@code fragment} に対応するヒントワードセットの {@code URI}
     */
    public static URI getLibraryHCFHintWordSetURI(final String fragment) {
        return URI.createURI(HCF_HINT_WORD_SETS_LIBRARY_URI).appendFragment(fragment);
    }

    /**
     * システム組み込みのヒントワードセットのリソースから {@code fragment} を持つヒントワードセットを返します。
     * 
     * @param objectContext
     *            リソースを扱う際のコンテキストオブジェクト
     * @param fragment
     *            取得するヒントワードセットの ID を示す URI フラグメント
     * @return {@code fragment} に対応するヒントワードセット
     */
    public static HCFHintWordSet getLibraryHCFHintWordSet(
            final EObject objectContext,
            final String fragment) {
        return (HCFHintWordSet) getLibraryObject(objectContext, fragment, HCF_HINT_WORD_SET);
    }

    /**
     * システム組み込みのヒントワードセットのリソースから {@code fragment} を持つヒントワードセットを一時的なリソース上にロードして返します。
     * 
     * @param fragment
     *            取得するヒントワードセットの ID を示す URI フラグメント
     * @return {@code fragment} に対応するヒントワードセット
     */
    public static HCFHintWordSet createLibraryHCFHintWordSet(final String fragment) {
        final ResourceSet resourceSet = new ResourceSetImpl();
        return getLibraryHCFHintWordSet(resourceSet, fragment);
    }

    /**
     * システム組み込みのヒントワードセットのリソースから {@code fragment} を持つヒントワードセットを指定された
     * {@code resourceSet} 上に展開します。
     * 
     * @param resourceSet
     *            ヒントワードセットの展開先
     * @param fragment
     *            取得するヒントワードセットの ID を示す URI フラグメント
     * @return {@code fragment} に対応するヒントワードセット
     */
    public static HCFHintWordSet getLibraryHCFHintWordSet(
            final ResourceSet resourceSet,
            final String fragment) {
        final URI uri = getLibraryHCFHintWordSetURI(fragment);
        final EObject eObject = resourceSet.getEObject(uri, true);
        return (HCFHintWordSet) eObject;
    }

    /**
     * システム組み込みのすべてのヒントワードセットを一時的なリソース上に展開します。
     * 
     * @return システム組み込みのすべてのヒントワードセットのリスト
     */
    public static EList<HCFHintWordSet> createLibraryDefaultHCFHintWordSets() {
        final ResourceSet resourceSet = new ResourceSetImpl();
        return getLibraryDefaultHCFHintWordSets(resourceSet);
    }

    /**
     * システム組み込みのすべてのヒントワードセットを指定された {@code resourceSet} 上に展開します。
     * 
     * @return システム組み込みのすべてのヒントワードセットのリスト
     */
    public static EList<HCFHintWordSet> getLibraryDefaultHCFHintWordSets(
            final ResourceSet resourceSet) {
        final URI uri = getLibraryHCFHintWordSetURI("_0");
        final EObject eObject = resourceSet.getEObject(uri, true);
        final STPAAnalysis stpaAnalysis = (STPAAnalysis) eObject;
        return stpaAnalysis.getHCFHintWordSets();
    }

    static EObject getLibraryObject(final EObject objectContext, final String fragment,
            final EClass eClass) {
        final InternalEObject eObject = (InternalEObject) EcoreUtil.create(eClass);
        eObject.eSetProxyURI(getLibraryHCFHintWordSetURI(fragment));
        return EcoreUtil.resolve(eObject, objectContext);
    }

}
